﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Services.Hive.DataTransfer;

namespace HeuristicLab.Services.Hive {
  public class RoundRobinTaskScheduler : ITaskScheduler {
    private IHiveDao dao {
      get { return ServiceLocator.Instance.HiveDao; }
    }

    public IEnumerable<Task> Schedule(IEnumerable<Task> tasks, int count = 1) {
      var userPriorities = dao.GetUserPriorities(x => true).ToArray();

      var jobs = new List<Job>();
      foreach (var userPriority in userPriorities) {
        var closureUserPriority = userPriority;
        jobs.AddRange(dao.GetJobs(x => x.OwnerUserId == closureUserPriority.UserId));
      }

      var taskJobRelations = tasks.Join(jobs,
        task => task.JobId,
        job => job.Id,
        (task, job) => new { Task = task, Job = job })
        .ToDictionary(k => k, v => false);

      var scheduledTasks = new List<Task>();
      int priorityIndex = 0;
      if (count == 0 || count > taskJobRelations.Count) count = taskJobRelations.Count;
      for (int i = 0; i < count; i++) {
        var defaultEntry = taskJobRelations.First(x => !x.Value); // search first task which is not included yet
        var priorityEntries = taskJobRelations.Where(x => !x.Value && x.Key.Job.OwnerUserId == userPriorities[priorityIndex].UserId).ToArray(); // search for tasks with desired user priority
        while (!priorityEntries.Any() && priorityIndex < priorityEntries.Length)
          priorityEntries = taskJobRelations.Where(x => !x.Value && x.Key.Job.OwnerUserId == userPriorities[++priorityIndex].UserId).ToArray();
        if (priorityEntries.Any()) { // tasks with desired user priority found
          var priorityEntry = priorityEntries.OrderByDescending(x => x.Key.Task.Priority).ThenBy(x => x.Key.Task.DateCreated).First();
          if (defaultEntry.Key.Task.Priority <= priorityEntry.Key.Task.Priority) {
            taskJobRelations[priorityEntry.Key] = true;
            scheduledTasks.Add(priorityEntry.Key.Task);
            priorityIndex++;
          } else { // there are other tasks with higher priority
            taskJobRelations[defaultEntry.Key] = true;
            scheduledTasks.Add(defaultEntry.Key.Task);
          }
        } else {
          taskJobRelations[defaultEntry.Key] = true;
          scheduledTasks.Add(defaultEntry.Key.Task);
        }
      }

      // dequeue and enqueue priorities [0, priorityIndex), i.e. {4,3,6,2,5,1} and priorityIndex = 3 -> {2,5,1,4,3,6}
      // i don't know how to solve this yet ... (race conditions, locking?)

      return scheduledTasks;
    }
  }
}
